/// VSOS APPLICATION FOR VS1005G DEVELOPER BOARD

/// MP3/WAV Player VSOS/Model-View-Controller Example Application for VS1005G
/// Uses the automatic ui generator kernel module AU.SYS

/// Compiling this software creates an .APP file, copy it 
/// to the VS1005 SD card to run it from the VSOS.
/// If you give it name "INIT.APP" then it will be run automatically by the kernel.

/// Please visit www.vsdsp-forum.com for free support and discussion!

/// NOTE: This is a special release that supports MP3/WAV playback although VSOS
/// doesn't yet fully support this at the time of writing (130322). Because of this,
/// parts of the operating system, namely decodeAudio.c and .h, and libcodecwav.a,
/// are integrated to this application. Later this functionality will be included
/// in VSOS, and a playback application will become simpler.

#include <vstypes.h>
#include <vo_stdio.h>
#include <vsos.h>
#include <sysmemory.h>
#include <vsostasks.h>
#include <timers.h>
#include "uimessages.h"
#include "mp3model.h" // The MP3 decoder Model
#include "mp3ui.h" // The View and Controller of MP3 player
 
__x s_int16 xmemPool[7500]; // These are the minimum amounts of Malloc Area for MP3 decoding to work.
__y s_int16 ymemPool[3800]; // These are the minimum amounts of Malloc Area for MP3 decoding to work.

//__x s_int16 xmemPool[6000+8765]; //__y s_int16 ymemPool[4000+8765]; // Roughly this much is maximally available

auto void InitMem(void) {
  __InitMemAlloc(xmemPool, sizeof(xmemPool),
		 ymemPool, sizeof(ymemPool));
}

int main(void) {
  // Because we're not sure how we are called, make sure that interrupts are enabled
  Enable(); Enable();
  InitMem(); // Setup memory areas for malloc
  
  // Start the MP3 decoder task to run in the background
  StartTask(TASK_DECODER, ModelTask); //in mp3model.c
  Yield(); // Release the CPU to allow the MP3 decoder task to initialize

  Repaint(); //Start the user interface

  // Inform the MP3 player model that we are listening to 
  // messages in function "PlayerCallback"
  ModelCallback(-1, UIMSG_VSOS_SET_CALLBACK_FUNCTION,
			(u_int32)PlayerCallback);
			
  // Ask the MP3 player model to send us a list of accepted messages		
  ModelCallback(-1, UIMSG_VSOS_GET_MSG_LIST, 0);
  Yield(); // Release the CPU to allow the MP3 model and PlayerCallback to process the message

  
  // Send a "Button Press: First" message to the MP3 player model
  // so that it would start decoding the first file on the disk
  ModelCallback(-1, UIMSG_BUT_FIRST, 1); 
  Yield(); //Release the CPU to allow the MP3 model to process the message

  while (pSysTasks[TASK_DECODER].task.tc_State && pSysTasks[TASK_DECODER].task.tc_State != TS_REMOVED) {
  	Delay(1); //Release the CPU for 1 millisecond to process other tasks or sleep to save power
	UiHandler(); //Call the user interface handler (in mp3ui.c) to process the touch screen etc
  }
  
	return EXIT_SUCCESS;
}
