/*

  MP3 Decoder Model (C) 2012-07-13 VLSI Solution Oy
  This model can play back MP3 files on VS1005g.

  For more details, have a look at mp3model.h

*/
#include <vo_stdio.h>	// MUST be first include!!!
#include <string.h>
#include <vo_fat.h>
#include <vsos.h>
#include <codecmpg.h>
#include <codec.h>
#include <timers.h>
#include <exec.h>
#include <swap.h>
#include <ctype.h>
#include <vs1005g.h>
#include "decodeAudio.h"
#include "uimessages.h"
#include "vo_fatdirops.h"
#include "mp3model.h"
#include "lcd.h"



AUDIO_DECODER *auDec = NULL;

/* Function pointer to caller callback function. */
void (*playerCallbackFunction)(s_int16 index, u_int16 message, u_int32 value)
  = NULL;


/* ID3v3 reading routines */
s_int32 GetID3Size7(FILE *fp) {
  int i;
  u_int32 res = 0;
  for (i=0; i<4; i++) {
    res <<= 7;
    res |= fgetc(fp) & 0x7F;
  }
  return res;
}

s_int32 GetID3Size8(FILE *fp) {
  int i;
  u_int32 res = 0;
  for (i=0; i<4; i++) {
    res <<= 8;
    res |= fgetc(fp);
  }
  return res;
}

#define ID3F_UNSYNCHRONIZED	0x80
#define ID3F_EXTENDED		0x40
#define ID3F_EXPERIMENTAL	0x20

struct ID3Tags {
  u_int16 uiMessage;
  char tagStr[4];
};

const struct ID3Tags id3Tags[] = {
  {UIMSG_TEXT_SONG_NAME, "TIT2"},
  {UIMSG_TEXT_ALBUM_NAME, "TALB"},
  {UIMSG_TEXT_ARTIST_NAME, "TPE1"},
  {UIMSG_TEXT_YEAR, "TYER"},
  {UIMSG_VSOS_END_OF_LIST, ""}
};

/* Must be at least 158 words, requirement by DecodeID3 */
#define MAX_TMP_STR_LEN 256
char currTmpStr[MAX_TMP_STR_LEN];


/*
  Decodes ID3v1 and ID3v2 tags.
*/
auto void DecodeID3(FILE *fp) {
#if 0
  printf("DecodeID3()\n");
#endif

  /* Look for ID3v2 */
  if (fgetc(fp) == 'I' && fgetc(fp) == 'D' && fgetc(fp) == '3' &&
      fgetc(fp) == 0x03 && fgetc(fp) == 0x00) {
    s_int32 id3Left;
    u_int16 flags = fgetc(fp);
    s_int16 i;
    id3Left = GetID3Size7(fp);
#if 0
    printf("flags = 0x%x, size = 0x%lx\n", flags, id3Left);
#endif
    if (flags & ID3F_EXTENDED) {
      u_int16 extSize = GetID3Size8(fp);
      fseek(fp, extSize, SEEK_CUR);
      id3Left -= extSize;
    }
    while (id3Left > 0) {
      char frameId[4];
      s_int32 frameSize;
      u_int16 frameFlags;
      u_int16 thereIsFrameId = 0;

#if 0
      printf("POS 0x%lx = %ld, 0x%lx left\n",
	     ftell(fp), ftell(fp), id3Left);
#endif

      for (i=0; i<4 && id3Left > 0; i++) {
	thereIsFrameId |= (frameId[i] = fgetc(fp));
	id3Left--;
      }
      if (thereIsFrameId && id3Left >= 6) {
	frameSize = GetID3Size8(fp);
	frameFlags = (u_int16)fgetc(fp) << 8;
	frameFlags |= fgetc(fp);
	id3Left -= 6;
	if (id3Left < frameSize)
	  frameSize = id3Left;
	id3Left -= frameSize;
#if 0
	printf("  Frame %c%c%c%c, size 0x%lx, flags 0x%04x\n",
	       frameId[0], frameId[1], frameId[2], frameId[3],
	       frameSize, frameFlags);
#endif
	i = 0;
	while (id3Tags[i].uiMessage != UIMSG_VSOS_END_OF_LIST) {
	  if (!memcmp(id3Tags[i].tagStr, frameId, 4)) {
	    u_int16 unicode = 0, skip = 0;
	    int j = 0;

	    /* Fundamentally incorrect way of handling unicode, but it
	       will show ASCII characters correctly. */

	    while (frameSize > 0 && j < MAX_TMP_STR_LEN) {
	      currTmpStr[j++] = fgetc(fp);
	      frameSize--;
	      if (unicode) {
		if (skip)
		  j--;
		skip ^= 1;
	      }
	      if (j == 3 && currTmpStr[0] == 1) {
		unicode = 2;
		skip = (currTmpStr[1] == 0xFE);
	      }
	    }
	    currTmpStr[j] = '\0';
#if 0
	    printf("Len %d, msg 0x%x, %s\n",
		   j, id3Tags[i].uiMessage, currTmpStr+1+unicode);
#endif
	    if (playerCallbackFunction) {
	      playerCallbackFunction(-1, id3Tags[i].uiMessage,
				     (u_int32)currTmpStr+1+unicode);
	    }
	  }
	  i++;
	}

	fseek(fp, frameSize, SEEK_CUR);
#if 0
	printf("  pos 0x%lx = %ld, 0x%lx left of id3\n",
	       ftell(fp), ftell(fp), id3Left);
	Delay(1000);
#endif
      } else {
#if 0
	printf("  No frame: ftell 0x%lx = %ld, id3Left 0x%lx = %ld\n",
	       ftell(fp), ftell(fp), id3Left, id3Left);
#endif
	fseek(fp, id3Left, SEEK_CUR);
	id3Left = 0;
      }
#if 0
      Delay(1000);
#endif
    }
#if 0
    printf("Done, left 0x%lx = %ld, ftell() = 0x%lx = %ld\n",
	   id3Left, id3Left, ftell(fp), ftell(fp));
#endif
  } else {
    /* Check for ID3V1 tag */
    s_int16 i;
    char *c = currTmpStr+MAX_TMP_STR_LEN-128;

    fseek(fp, -128, SEEK_END);
    for (i=0; i<128; i++) {
      *c++ = fgetc(fp);
    }
    c = currTmpStr+MAX_TMP_STR_LEN-128;
#if 0
    printf("Seek for ID3v1 %x %x %x\n", c[0], c[1], c[2]);
#endif
    if (!strncmp(c, "TAG", 3) && playerCallbackFunction) {
      /* Found ID3v1 tag */
      c += 3;
      currTmpStr[30] = '\0';

      memcpy(currTmpStr, c, 30);
      c += 30;
      playerCallbackFunction(-1, UIMSG_TEXT_SONG_NAME, (u_int32)currTmpStr);

      memcpy(currTmpStr, c, 30);
      c += 30;
      playerCallbackFunction(-1, UIMSG_TEXT_ARTIST_NAME, (u_int32)currTmpStr);

      memcpy(currTmpStr, c, 30);
      c += 30;
      playerCallbackFunction(-1, UIMSG_TEXT_ALBUM_NAME, (u_int32)currTmpStr);

      currTmpStr[4] = '\0';
      memcpy(currTmpStr, c, 4);
      playerCallbackFunction(-1, UIMSG_TEXT_YEAR, (u_int32)currTmpStr);
      
    }
    fseek(fp, 0, SEEK_SET);
  }
}





/* List of user interface message items this model can handle. */
UiMessageListItem __y playerUiMessageListItem[] = {
	{UIMSG_HINT,"LAY=1/DIV=40"},//Suggestion of screen layout strategy
	{UIMSG_VSOS_SET_CALLBACK_FUNCTION,"Set callback"},
	{UIMSG_VSOS_GET_MSG_LIST,"Get MSG list"},
	
	// Model's input user interface 
	// (the messages for controlling the operation of the model 
	// which are relevant for automatic generation of the user interface)
	{UIMSG_INPUTS, "File Controls"}, //Start of input category "File Controls"
	{UIMSG_HINT,"MIN=0/MAX=10"},//Set slide control limits
	{UIMSG_BUT_FIRST,"First"},
	{UIMSG_BUT_PREVIOUS,"Previous"},
	{UIMSG_BUT_NEXT,"Next"},
	{UIMSG_BUT_LAST,"Last"},
	{UIMSG_TEXT_OPEN_FILE,"Open file"},
	
	{UIMSG_INPUTS,"Play Controls"}, //Start of input category "Play Controls"
	{UIMSG_BUT_RESTART,"Restart"},
	{UIMSG_BUT_STOP,"Stop"},
	{UIMSG_BUT_PAUSE,"Pause"},
	{UIMSG_BUT_PAUSE_TOGGLE,"Pause toggle"},
	{UIMSG_BOOL_FAST_FORWARD,"Fast forward"},
	{UIMSG_BUT_CLOSE_DECODER,"Close decoder"},
	{UIMSG_BUT_SAVE_POS,"Save Position"},
	{UIMSG_BUT_RESTORE_POS, "Restore Position"},
	
	// Model's output user interface 
	// (those status messages which the model generates,
	// which are relevant for automatic generation of the user interface)
	{UIMSG_OUTPUTS, "Song Info"}, //Start of output category "Song Info"
	{UIMSG_TEXT_SHORT_FILE_NAME, ""},
	{UIMSG_USE_SAME_POSITION, NULL},{UIMSG_TEXT_LONG_FILE_NAME,""},
	{UIMSG_TEXT_ARTIST_NAME, "Artist: "},
	{UIMSG_TEXT_SONG_NAME, "Song:   "},
	{UIMSG_TEXT_ALBUM_NAME, "Album:  "},
	{UIMSG_USE_SAME_LINE, NULL}, {UIMSG_TEXT_YEAR, "Year:   "},
	{UIMSG_U32_PLAY_FILE_PERCENT, "Play Time: "},
	{UIMSG_USE_SAME_LINE, NULL}, {UIMSG_U32_PLAY_TIME_SECONDS, " "},
	{UIMSG_VSOS_END_OF_LIST,NULL}
};



u_int16 shouldSendUiMessageList = 0;

/* Send the UI Message list */
auto void SendUiMessageList(void) {
  if (playerCallbackFunction)
    playerCallbackFunction(-1, UIMSG_VSOS_MSG_LIST,
			   (u_int32)playerUiMessageListItem);
  shouldSendUiMessageList = 0;
}



/* States of the player state machine */
enum PlayerStates {
  psIdle,
  psOpenFile,
  psPlaying,
  psPaused,
  psClearPause,
  psCloseFile,
  psEnd
};


/* Store space for different player signals */
struct PLAYER_SIGNALS {
  enum OpenFileMode openFile;
  s_int16 stop,
    pauseToggle, pauseOn, pauseOff, closeDecoder;
} pSig = {0};


/* Temporary storage for file names */
#define FILE_NAME_CHARS 256
char fileName[FILE_NAME_CHARS]="";

u_int32 lastSecond=0;
u_int16 lastPause=-1;





#include <codMDec.h>
#if 1
#include "layers_tiny.h"
#else
#include <layers.h>
#endif

u_int32 savePos = 0;
u_int32 blockPos[3] = {0, 0, 0};
enum SavePosOp {
  spoNone,
  spoSave,
  spoRestore
} savePosOp = spoNone;
/*
  This is a codec services callback function, called whenever audio
  samples are being output by the MP3 decoder during
  DecodeAudio(auDec->cod, &auDec->cs, &eStr);

  This way we get control every now and then even when decoding audio,
  and can e.g. send playback time related and other status messages.
*/
void CodecServicesCallback(AUDIO_DECODER *auDec, u_int16 samples) {
  static u_int32 totSamples = 0;
  u_int16 frame = (int)(totSamples/576)+1;
  totSamples += samples;

#if 1
  Forbid();
  if (auDec && auDec->inFp) {
    {
      u_int32 t = ftell(auDec->inFp);
      if (t != blockPos[2]) {
	/* Keep three last blockpositions in memory */
	blockPos[0] = blockPos[1];
	blockPos[1] = blockPos[2];
	blockPos[2] = t;
      }
    }
    if (savePosOp == spoSave) {
      savePos = blockPos[0];
    } else if (savePosOp == spoRestore) {
      struct CodecMpg *c;
      fseek(auDec->inFp, savePos, SEEK_SET);
      c = (struct CodecMpg *)auDec->cod;
      // Discard reservoir after jump
      if (c) {
#if 0
	printf("RESLEN *%p* = *%d*\n",
	       &(c->layer123y->fr.reservoirLen),
	       c->layer123y->fr.reservoirLen);
#endif
	c->layer123y->fr.reservoirLen = 0;
      }

      auDec->cs.fileLeft = auDec->cs.fileSize - savePos;
    }
  }
  savePosOp = spoNone;
  Permit();
#endif

  
  if (playerCallbackFunction) {
    if (auDec->cs.playTimeSeconds != lastSecond) {
      u_int16 perCent;
      lastSecond = auDec->cs.playTimeSeconds;
      playerCallbackFunction(-1, UIMSG_U32_PLAY_TIME_SECONDS, lastSecond);
      if (auDec->cs.fileLeft != 0xFFFFFFFFU) {
		perCent = 100.9-(auDec->cs.fileLeft*100.8/auDec->cs.fileSize);
		if (perCent > 100)
		  perCent = 100;
		playerCallbackFunction(-1, UIMSG_U32_PLAY_FILE_PERCENT, perCent);
	  }
    }
    if (auDec->pause != lastPause) {
      lastPause = auDec->pause;
      playerCallbackFunction(-1, UIMSG_S16_PLAYING, lastPause ?
			     UIM_PLAYING_PAUSED : UIM_PLAYING_PLAYING);
    }
    if (shouldSendUiMessageList)
      SendUiMessageList();
  }
  
	if (pSig.closeDecoder) {
		auDec->cs.cancel = 1;
	}
}




/*
  This callback function gets called by the MP3 player View / Controller
  part (UI).

  Because the standard Decoder is built in such a way that Decode() functions
  doesn't end execution, we have to insert a Cancel command for all the
  user operations that require action from our MP3 model while doing playback.
  Although it makes the state machine operations less clear, we need to
  put some of the state machine's operations here.
*/
void ModelCallback(s_int16 index, u_int16 message, u_int32 value) {
  struct Codec *cod;
  int shouldCancel = 0;

  /* Disable multitasking for the duration of this function. */
  Forbid();
  cod = (auDec) ? auDec->cod : NULL;

  switch (message) {
  case UIMSG_VSOS_SET_CALLBACK_FUNCTION:
    playerCallbackFunction = (void *)value;
    break;
  case UIMSG_VSOS_GET_MSG_LIST:
    shouldSendUiMessageList = 1;
    break;
  case UIMSG_TEXT_OPEN_FILE:
    pSig.openFile = ofmName;
    shouldCancel = 1;
    strcpy(fileName, (char *)value);
    break;
  case UIMSG_BUT_FIRST:
    pSig.openFile = ofmFirst;
    shouldCancel = 1;
    break;
  case UIMSG_BUT_NEXT:
    pSig.openFile = ofmNext;
    shouldCancel = 1;
    break;
  case UIMSG_BUT_PREVIOUS:
    pSig.openFile = ofmPrev;
    shouldCancel = 1;
    break;
  case UIMSG_BUT_LAST:
    pSig.openFile = ofmLast;
    shouldCancel = 1;
    break;
  case UIMSG_BUT_RESTART:
    pSig.openFile = ofmSame;
    shouldCancel = 1;
    break;
  case UIMSG_BUT_STOP:
    shouldCancel = 1;
    pSig.stop = value;
    break;
  case UIMSG_BUT_PAUSE:
    if (auDec) {
      auDec->pause = value;
    }
    break;
  case UIMSG_BUT_PAUSE_TOGGLE:
    if (auDec) {
      auDec->pause = !(auDec->pause);
    }
    break;
  case UIMSG_BOOL_FAST_FORWARD:
    if (cod) {
      cod->cs->fastForward = value ? 8 : 1;
    }
    break;
  case UIMSG_BUT_CLOSE_DECODER:
    pSig.closeDecoder = value;
    break;
  case UIMSG_BUT_SAVE_POS:
    savePosOp = spoSave;
    break;
  case UIMSG_BUT_RESTORE_POS:
    savePosOp = spoRestore;
    break;
  }
  if (shouldCancel && cod)
    cod->cs->cancel = 1;	

  Permit();
}




/*
  This is our MP3 model.

  It interprets messages that were received from the UI, the acts
  accordingly.

  At the moment there is no function to change directories, so all
  files are played from the main directory of the SD card. However,
  this will change on subsequent releases.

  Also, this decoder still contains lots of #if 0'd test code. It
  will be cleaned when a more final release is published.
*/
void ModelTask(void) {
  int i=0;
  //  static struct CodecServices cs = {0};
  const char *eStr;
  FILE *fp = NULL;
  enum PlayerStates playerState = psIdle;

  InitDirectory(&fp);


  /* Main loop */
  while (playerState != psEnd) {

    switch (playerState) {

    case psIdle:
      if (pSig.closeDecoder) {
	playerState = psEnd;
      } else if (pSig.openFile) {
	playerState = psOpenFile;
	if (pSig.openFile < ofmNone) {
	  static s_int16 fNum;
	  char *nameP = fileName;
	  if (!fileName[0]) {
	    fileName[0] = 'S';
	    fileName[1] = ':';
	    nameP = fileName+2;
	  }
	  fNum = FindFile(fp, nameP, FILE_NAME_CHARS, "mp3,wav,ima",
			  pSig.openFile, etFile);
#if 0
	  printf("  -> fNum=%d, nam=\"%s\"\n", fNum, fileName);
#endif
	  /* Activating this piece of code for some reason makes the
	     software trip over. Must be an unknown LCC bug. */
	  if (playerCallbackFunction) {
	    playerCallbackFunction(-1, UIMSG_S16_SONG_NUMBER, fNum);
	  }
	  if (fNum < 0) {
	    fileName[0] = '\0';
	    playerState = psIdle;
	    if (playerCallbackFunction) {
	      playerCallbackFunction(-1, UIMSG_TEXT_ERROR_MSG,
				     (u_int32)"File not found");
	    }
	  }
	}
	pSig.openFile = ofmNone;
      } else if (pSig.stop) {
	pSig.stop = 0;
      } else {
	/* Nothing to see here. Don't hog CPU. */
	Delay(10);
      }
      break;



    case psOpenFile:
      {
	FILE *inFp = fopen(fileName, "rb");
	Forbid();
	if (inFp) {
	  auDec = CreateAudioDecoder(inFp, stdaudioout, CodecServicesCallback,
				     auDecFGuess);
	}
	Permit();
	if (auDec) {
	  if (playerCallbackFunction) {
	    playerCallbackFunction(-1, UIMSG_TEXT_SHORT_FILE_NAME,
				   (u_int16)(fileName));
	    playerCallbackFunction(-1, UIMSG_TEXT_LONG_FILE_NAME,
				   (u_int16)(fileName+strlen(fileName)+1));
	  }
	

	  if (auDec->cs.fileLeft != 0xFFFFFFFFU) {
		DecodeID3(auDec->inFp);
	  }
	  playerState = psPlaying;
	} else {
	  if (inFp)
	    fclose(inFp);
	  if (playerCallbackFunction) {
	    playerCallbackFunction(-1, UIMSG_TEXT_ERROR_MSG,
				   (u_int32)"Cannot open audio");
	  }
	  fileName[0] = '\0';
	  playerState = psIdle;
	}
      }
      break;


    case psPlaying:
      if (pSig.closeDecoder) {
	playerState = psCloseFile;
      } else if (pSig.stop) {
	playerState = psCloseFile;
	pSig.stop = 0;
      } else if (pSig.openFile) {
	playerState = psCloseFile;
      } else if (auDec->cod->cs->fileLeft <= 0) {
	if (playerCallbackFunction)
	  playerCallbackFunction(-1, UIMSG_BUT_END_OF_SONG, 1);
	playerState = psCloseFile;
      } else {
	lastSecond = 12345;	// Will force time update
	DecodeAudio(auDec, &eStr);
	lastPause = -1;
	playerCallbackFunction(-1, UIMSG_S16_PLAYING, UIM_PLAYING_STOPPED);
	playerState = psCloseFile;
      }
      auDec->cod->cs->cancel = 0;
      break;



    case psCloseFile:
      if (auDec->inFp)
	fclose(auDec->inFp);
      DeleteAudioDecoder(auDec);
      auDec = NULL;
      fileName[0] = '\0';
      playerState = psIdle;
      break;


    default:
      printf("Internal error #1: playerState %d!\n",
	     playerState);
      playerState = psEnd;
      break;

    } /* switch(playerState) */

    if (shouldSendUiMessageList)
      SendUiMessageList();
  } /* while (playerState != psEnd) */



 end:
  EndDirectory(&fp);
  if (playerCallbackFunction) {
    playerCallbackFunction(-1, UIMSG_BUT_DECODER_CLOSED, 1);
  }
}
