/// \file mp3ui.c User interface for an MP3 player, which uses the Automatic User interface (uimaker)
/// The automatic user interface is implemented as a kernel module VSOS2GAU.SYS which can be changed
/// on the SD card if the user interface needs to be changed.
/// \author Panu-Kristian Poiksalo, VLSI Solution Oy 2012

#include <vstypes.h>
#include <vsos.h>
#include <string.h>
#include <vo_stdio.h>
#include "mp3ui.h"
#include "lcd.h"
#include "vs1005g.h"
#include "mp3model.h"
#include <exec.h>
#include <uimaker.h>
#include <timers.h>

const char *playerStateText[] = {"Stopped","Paused","Playing","Closed"};

//UiMessageListItem __y *uiMessageList = NULL;
StdButton myBut[20]; 
UiInfo ui = {myBut,ModelCallback,playerStateText};

// Application's repaint method should completely draw the application's screen from scratch.
void Repaint(void) {
	// Repaint the console frame
	console.Ioctl(&console, IOCTL_START_FRAME, "VS1005G VSOS/MVC/Generic UI MP3 Demo");
	
	memset(myBut,0,sizeof(myBut));
	
	// Generate automatically a set of buttons based on the model's list of button inputs
	 
	HighlightText();
	printf("VSOS MP3 Player Example Application\n");
	NormalText();
	printf("Uses Generic User Interface S:SYS/AU.SYS\n");	
	
	UserInterfaceCreate(&ui);
}
 

/// This is the receiver for UIMessages from the MP3 player model.
void *PlayerCallback(s_int16 index, u_int16 message, u_int32 value) {
	Forbid();
	ui.lastMessage->index = index;
	ui.lastMessage->message = message;
	ui.lastMessage->value = value;
	UserInterfaceSendMessage(&ui);
	Permit();	
}


// Call application's UiHandler method to give a chance for the buttons to be pressed
void UiHandler(void) {	
	u_int16 buttonResult = UserInterfaceGetEvent(&ui);
	if (buttonResult) {
		//printf("Button %04x %s\n",buttonResult,lastButtonPressed->caption);
		ModelCallback(-1, buttonResult, (u_int32)1);
	}
	
}
 